#include "ehs/io/socket/Comms.h"
#include "ehs/io/socket/CommsSystem.h"
#include "ehs/io/socket/Endpoint.h"
#include "ehs/Encryption.h"

namespace ehs
{
	const Version Comms::version(1, 0, 0);
	const UInt_64 Comms::internalSys = Str_8::Hash_64("Internal");
	const UInt_64 Comms::connectOp = Str_8::Hash_64("Connect");
	const UInt_64 Comms::connectedOp = Str_8::Hash_64("Connected");
	const UInt_64 Comms::rejectedOp = Str_8::Hash_64("Rejected");
	const UInt_64 Comms::disconnectOp = Str_8::Hash_64("Disconnect");
	const UInt_64 Comms::disconnectedOp = Str_8::Hash_64("Disconnected");
	const UInt_64 Comms::statusUpdateOp = Str_8::Hash_64("StatusUpdate");
	const UInt_64 Comms::pingOp = Str_8::Hash_64("Ping");
	const UInt_64 Comms::pongOp = Str_8::Hash_64("Pong");
	const UInt_64 Comms::latencyOp = Str_8::Hash_64("Latency");
	const UInt_64 Comms::receivedOp = Str_8::Hash_64("Received");

	Comms::~Comms()
	{
		UnInitialize();
	}

	Comms::Comms()
		: udp(AddrType::IPV6), appVer(0, 0, 0),
		disposition(EndDisp::UNKNOWN), dropPackets(false), hashId(0), buffer(nullptr), bufferSize(0),
		maxEndpoints(0), lastTSC(0), delta(0.0f), maxTimeout(5.0f), resendRate(0.5f), connectedCb(nullptr),
		activeCb(nullptr), disconnectedCb(nullptr)
	{
		AddType("Socket");
	}

	Comms::Comms(const Version& ver, const EndDisp disposition, const Str_8& id, const UInt_64 maxEndpoints)
		: udp(AddrType::IPV6), appVer(ver), disposition(disposition), dropPackets(false), hashId(id.Hash_64()), id(id),
		buffer(nullptr), bufferSize(0), maxEndpoints(maxEndpoints), lastTSC(CPU::GetTSC()), delta(0.0f),
		maxTimeout(5.0f), resendRate(0.5f), connectedCb(nullptr), activeCb(nullptr), disconnectedCb(nullptr)
	{
	}

	Comms::Comms(const Comms& sock)
		: BaseObj(sock), udp(sock.udp), appVer(sock.appVer), disposition(sock.disposition),
		dropPackets(sock.dropPackets), hashId(sock.hashId), id(sock.id), buffer(nullptr), bufferSize(0),
		maxEndpoints(sock.maxEndpoints), lastTSC(CPU::GetTSC()), delta(0.0f), maxTimeout(sock.maxTimeout),
		resendRate(sock.resendRate), connectedCb(nullptr), activeCb(nullptr), disconnectedCb(nullptr)
	{
		AddType("Socket");
	}

	Comms& Comms::operator=(const Comms& sock)
	{
		if (this == &sock)
			return *this;

		BaseObj::operator=(sock);

		udp = sock.udp;
		appVer = sock.appVer;
		disposition = sock.disposition;
		dropPackets = sock.dropPackets;
		hashId = sock.hashId;
		id = sock.id;
		buffer = nullptr;
		bufferSize = 0;
		systems = Array<CommsSystem*>();
		endpoints = Vector<Endpoint*>();
		maxEndpoints = sock.maxEndpoints;
        lastTSC = 0;
        delta = 0.0f;
		maxTimeout = sock.maxTimeout;
        resendRate = sock.resendRate;
		connectedCb = nullptr;
		activeCb = nullptr;
		disconnectedCb = nullptr;

		return *this;
	}

	void Comms::Initialize()
	{
		if (!udp.IsValid())
			return;

		if (udp.GetLocalAddressType() == AddrType::IPV4)
		{
			buffer = new Byte[EHS_IPV4_UDP_PAYLOAD];
			bufferSize = EHS_IPV4_UDP_PAYLOAD;
		}
		else if (udp.GetLocalAddressType() == AddrType::IPV6)
		{
			buffer = new Byte[EHS_IPV6_UDP_PAYLOAD];
			bufferSize = EHS_IPV6_UDP_PAYLOAD;
		}
	}

	void Comms::UnInitialize()
	{
		if (!udp.IsValid())
			return;

		delete[] buffer;
        buffer = nullptr;
        bufferSize = 0;

		Serializer payload(Endianness::LE);
		payload.Write<UInt_64>(0);

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetStatus() != Status::PENDING)
				endpoints[i]->Send(false, true, false, internalSys, disconnectOp, payload);

			delete endpoints[i];
		}
		endpoints.Clear();

		for (UInt_64 i = 0; i < systems.Size(); ++i)
			delete systems[i];
		systems.Clear();

		udp.Release();
	}

	void Comms::Bind(AddrType newType, const Str_8& newAddress, const UInt_16 newPort)
	{
		udp.Bind(newType, newAddress, newPort);
	}

	void Comms::Connect(AddrType rType, const Str_8& rAddress, const UInt_16 rPort)
	{
		if (!udp.IsValid())
			return;

		Serializer payload(Endianness::LE);
		payload.Write(CPU::GetArchitecture());
		payload.WriteStr(id);
		payload.WriteVersion(version);
		payload.WriteVersion(appVer);

		Endpoint* end = new Endpoint(this, rType, rAddress, rPort);
		end->Send(false, true, false, "Internal", "Connect", payload);

		endpoints.Push(end);
	}

	bool Comms::Disconnect(const EndDisp endDisp, const UInt_64 endHashId, const Str_8& msg)
	{
		if (!udp.IsValid())
			return false;

		Endpoint* end = GetEndpoint(endDisp, endHashId);
		if (!end)
			return false;

		Str_8 dcMsg = end->GetId() + " has been disconnected.";
		if (msg.Size())
			dcMsg += " Reason: " + msg;

		EHS_LOG_INT("Info", 0, dcMsg);

		Serializer<> payload(Endianness::LE);
		payload.WriteStr(msg);

		end->Send(false, true, false, internalSys, disconnectOp, payload);

		return true;
	}

	bool Comms::Disconnect(const EndDisp endDisp, const Str_8& endId, const Str_8& msg)
	{
		return Disconnect(endDisp, endId.Hash_64(), msg);
	}

	void Comms::Broadcast(const EndDisp endDisp, const Status endStatus, const bool deltaLocked, const bool encrypted,
						   const bool ensure, const UInt_64 sysHashId, const UInt_64 opHashId,
						   const Serializer<>& payload)
	{
		if (!udp.IsValid())
			return;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				endpoints[i]->Send(deltaLocked, encrypted, ensure, sysHashId, opHashId, payload);
		}
	}

	void Comms::Broadcast(const EndDisp endDisp, const Status endStatus, const bool deltaLocked, const bool encrypted,
						   const bool ensure, const Str_8& sysId, const Str_8& opId,
						   const Serializer<>& payload)
	{
		Broadcast(endDisp, endStatus, deltaLocked, encrypted, ensure, sysId.Hash_64(), opId.Hash_64(), payload);
	}

	void Comms::Poll()
	{
		if (udp.IsValid())
			return;

        UInt_64 newTSC = CPU::GetTSC();
        delta = (float)(newTSC - lastTSC) / (float)CPU::GetTSC_Freq();
        lastTSC = newTSC;

		AddrType rType = AddrType::IPV6;
		Str_8 rAddress;
		UInt_16 rPort = 0;

		UInt_16 received;

		while ((received = udp.Receive(&rType, &rAddress, &rPort, buffer, bufferSize)))
		{
			Serializer<> payload(Endianness::LE, buffer, received);

			bool encrypted = payload.Read<bool>();
			if (encrypted)
				Encryption::Encrypt_64(payload.Size() - payload.GetOffset(), &payload[payload.GetOffset()]);

            payload.SetOffset(0);

			Header header = payload.Read<Header>();

			if (!header.ensure && header.endpointId && header.system == internalSys && header.op == connectOp)
			{
				Architecture rArch = payload.Read<Architecture>();
				Str_8 rId = payload.ReadStr<Char_8, UInt_64>();

				Endpoint* end = new Endpoint(this, header.disposition, rArch, rId, rType, rAddress, rPort);
				end->SetStatus(Status::PENDING);

				Serializer sPayload(Endianness::LE);

				Version rVer = payload.ReadVersion();
				if (rVer != version)
				{
					sPayload.WriteStr<Char_8, UInt_64>("Your Event Horizon Socket Layer version " +
						Str_8::FromNum(rVer.major) + "." + Str_8::FromNum(rVer.minor) + "." + Str_8::FromNum(rVer.patch) +
						" does not match remote endpoint version " +
						Str_8::FromNum(version.major) + "." + Str_8::FromNum(version.minor) + "." + Str_8::FromNum(version.patch) +
						". Connection rejected.");

					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				Version rAppVer = payload.ReadVersion();
				if (rAppVer != appVer)
				{
					sPayload.WriteStr<Char_8, UInt_64>("Your application version " +
						Str_8::FromNum(rAppVer.major) + "." + Str_8::FromNum(rAppVer.minor) + "." + Str_8::FromNum(rAppVer.patch) +
						" does not match remote endpoint application version " +
						Str_8::FromNum(appVer.major) + "." + Str_8::FromNum(appVer.minor) + "." + Str_8::FromNum(appVer.patch) +
						". Connection rejected.");

					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				if (HasEndpoint(header.disposition, header.endpointId))
				{
					if (header.disposition == EndDisp::SERVICE)
					{
						sPayload.WriteStr<Char_8, UInt_64>(
								"The service, \"" + end->GetId() + "\", is taken. Connection rejected.");
						end->Send(false, true, false, internalSys, rejectedOp, sPayload);
						continue;
					}
					else
					{
						sPayload.WriteStr<Char_8, UInt_64>(
								"The username, \"" + end->GetId() + "\", is taken. Connection rejected.");
						end->Send(false, true, false, internalSys, rejectedOp, sPayload);
						continue;
					}
				}

				if (connectedCb && !connectedCb(this, end))
				{
					sPayload.WriteStr<Char_8, UInt_64>("Connection rejected.");
					end->Send(false, true, false, internalSys, rejectedOp, sPayload);
					continue;
				}

				endpoints.Push(end);

				sPayload.Write(CPU::GetArchitecture());
				sPayload.WriteStr(id);

				UInt_64 active = GetEndpointsCount(EndDisp::ENDPOINT, Status::ACTIVE);

				if (maxEndpoints && active >= maxEndpoints)
				{
					end->SetStatus(Status::IN_LOCAL_QUEUE);

					UpdateQueue(active);

					sPayload.Write(Status::IN_REMOTE_QUEUE);
					sPayload.Write(end->GetQueueSlot());

					EHS_LOG_INT("Info", 1, end->GetId() + " connected and is in queue slot " + end->GetQueueSlot() + ".");
				}
				else
				{
					end->SetStatus(Status::ACTIVE);

					if (activeCb)
						activeCb(this, end);

					sPayload.Write(Status::ACTIVE);
					sPayload.Write(0);

					EHS_LOG_INT("Info", 1, end->GetId() + " connected.");
				}

				end->Send(false, true, false, internalSys, connectedOp, sPayload);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == connectedOp)
			{
				Endpoint* end = GetEndpoint(rAddress, rPort);
				if (!end || end->GetStatus() != Status::PENDING)
					continue;

				Architecture arch = payload.Read<Architecture>();
				Str_8 endId = payload.ReadStr<Char_8, UInt_64>();

				*end = Endpoint(this, header.disposition, arch, endId, rType, rAddress, rPort);
				end->SetStatus(payload.Read<Status>());
				end->SetQueueSlot(payload.Read<UInt_64>());

				if (connectedCb)
					connectedCb(this, end);

				Str_8 msg = "Successfully connected to " + end->GetId();
				if (end->GetStatus() == Status::IN_REMOTE_QUEUE)
					msg += " and in queue slot " + Str_8::FromNum(end->GetQueueSlot()) + ".";
				else
					msg += ".";

				EHS_LOG_INT("Info", 2, msg);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == rejectedOp)
			{
				if (!RemoveEndpoint(rAddress, rPort))
					continue;

				Str_8 msg = payload.ReadStr<Char_8, UInt_64>();
				if (msg.Size())
					EHS_LOG_INT("Info", 3, msg);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == disconnectOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				end->Send(false, true, false, internalSys, disconnectedOp, {});

				if (disconnectedCb)
					disconnectedCb(this, end);

				Str_8 dcMsg;

				if (header.disposition == EndDisp::SERVICE)
					dcMsg = "You have been disconnected from, \"" + end->GetId() + "\".";
				else
					dcMsg = end->GetId() + " has disconnected.";

				Str_8 msg = payload.ReadStr<Char_8, UInt_64>();
				if (msg.Size())
					dcMsg += " Reason: " + msg;

				EHS_LOG_INT("Info", 4, dcMsg);

				RemoveEndpoint(header.disposition, end->GetHashId());

				UpdateQueue();
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == disconnectedOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				if (disconnectedCb)
					disconnectedCb(this, end);

				RemoveEndpoint(end);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == statusUpdateOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				Status newStatus = payload.Read<Status>();
				UInt_64 newSlot = payload.Read<UInt_64>();

				if (end->GetStatus() == Status::ACTIVE)
				{
					if (activeCb)
						activeCb(this, end);

					EHS_LOG_INT("Info", 5, "Your connection status to " + end->GetId() + " has now become active.");
				}
				else if (end->GetStatus() == Status::IN_REMOTE_QUEUE && newStatus == Status::IN_REMOTE_QUEUE)
				{
					EHS_LOG_INT("Info", 5, "Your queue slot for " + end->GetId() + " is now " + newSlot + ".");
				}

				end->SetStatus(newStatus);
				end->SetQueueSlot(newSlot);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == pingOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				end->SetDeltaRate(payload.Read<float>());
                end->Pong(delta);
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == pongOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				end->SetDeltaRate(payload.Read<float>());
                end->SendLatency();
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == latencyOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				end->SetLatency(payload.Read<float>());
			}
			else if (!header.ensure && header.endpointId && header.system == internalSys && header.op == receivedOp)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

                UInt_64 msgId = payload.Read<UInt_64>();
                UInt_64 fragment = payload.Read<UInt_64>();

				end->RemoveInsurance(msgId, fragment);
			}
			else if (header.endpointId)
			{
				Endpoint* end = GetEndpoint(header.disposition, header.endpointId);
				if (!end)
					continue;

				if (dropPackets && !header.ensure && header.id < end->GetNextRecvId())
                {
                    EHS_LOG_INT("Info", 6, "Old packet intentionally dropped.");
                    continue;
                }

                if (header.ensure)
                {
                    Serializer sPayload(Endianness::LE);
                    sPayload.Write(header.id);
                    sPayload.Write(header.fragment);

                    end->Send(false, true, false, internalSys, receivedOp, sPayload);
                }

				end->AddReceived(
					header,
					Serializer<>(Endianness::LE, &payload[payload.GetOffset()], payload.Size() - payload.GetOffset())
				);
			}
			else
			{
				EHS_LOG_INT("Info", 7, "Corrupted packet.");
			}
		}

		PollEndpoints();
	}

	bool Comms::IsInitialized() const
	{
		return udp.IsValid();
	}

	AddrType Comms::GetLocalAddressType() const
	{
		return udp.GetLocalAddressType();
	}

	Str_8 Comms::GetLocalAddress() const
	{
		return udp.GetLocalAddress();
	}

	UInt_16 Comms::GetLocalPort() const
	{
		return udp.GetLocalPort();
	}

	bool Comms::IsBound() const
	{
		return udp.IsBound();
	}

	Version Comms::GetVersion()
	{
		return version;
	}

	Version Comms::GetAppVersion() const
	{
		return appVer;
	}

	EndDisp Comms::GetDisposition() const
	{
		return disposition;
	}

	void Comms::EnableDropPackets(const bool enable)
	{
		dropPackets = enable;
	}

	bool Comms::IsDropPacketsEnabled() const
	{
		return dropPackets;
	}

	Str_8 Comms::GetId() const
	{
		return id;
	}

	UInt_64 Comms::GetHashId() const
	{
		return hashId;
	}

	bool Comms::HasSystem(const UInt_64 sysHashId) const
	{
		if (internalSys == sysHashId)
			return true;

		for (UInt_64 i = 0; i < systems.Size(); ++i)
			if (systems[i]->GetHashId() == sysHashId)
				return true;

		return false;
	}

	bool Comms::HasSystem(const Str_8& sysId) const
	{
		return HasSystem(sysId.Hash_64());
	}

	bool Comms::AddSystem(CommsSystem* sys)
	{
		if (HasSystem(sys->GetHashId()))
			return false;

		systems.Push(sys);

		return true;
	}

	CommsSystem* Comms::GetSystem(const UInt_64 sysHashId)
	{
		for (UInt_64 i = 0; i < systems.Size(); ++i)
			if (systems[i]->GetHashId() == sysHashId)
				return systems[i];

		return nullptr;
	}

	CommsSystem* Comms::GetSystem(const Str_8& sysId)
	{
		return GetSystem(sysId.Hash_64());
	}

	bool Comms::HasEndpoint(const EndDisp endDisp, const Status endStatus, const UInt_64 endHashId) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (endpoints[i]->GetHashId() == endHashId)
				return true;
		}

		return false;
	}

	bool Comms::HasEndpoint(const EndDisp endDisp, const Status endStatus, const Str_8& endId) const
	{
		return HasEndpoint(endDisp, endStatus, endId.Hash_64());
	}

	bool Comms::HasEndpoint(const EndDisp endDisp, const UInt_64 endHashId) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetHashId() == endHashId)
				return true;
		}

		return false;
	}

	bool Comms::HasEndpoint(const EndDisp endDisp, const Str_8& endId) const
	{
		return HasEndpoint(endDisp, endId.Hash_64());
	}

	bool Comms::HasEndpoint(const Str_8& rAddress, const UInt_16 rPort) const
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
				return true;

		return false;
	}

	Endpoint* Comms::GetEndpoint(const EndDisp endDisp, const Status endStatus, const UInt_64 endHashId)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() != endStatus)
				continue;

			if (endpoints[i]->GetHashId() == endHashId)
				return endpoints[i];
		}

		return nullptr;
	}

	Endpoint* Comms::GetEndpoint(const EndDisp endDisp, const Status endStatus, const Str_8& endId)
	{
		return GetEndpoint(endDisp, endStatus, endId.Hash_64());
	}

	Endpoint* Comms::GetEndpoint(const EndDisp endDisp, const UInt_64 endHashId)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetHashId() == endHashId)
				return endpoints[i];
		}

		return nullptr;
	}

	Endpoint* Comms::GetEndpoint(const EndDisp endDisp, const Str_8& endId)
	{
		return GetEndpoint(endDisp, endId.Hash_64());
	}

	Endpoint* Comms::GetEndpoint(const Str_8& rAddress, const UInt_16 rPort)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
				return endpoints[i];

		return nullptr;
	}

	Array<Endpoint*> Comms::GetEndpoints(const EndDisp endDisp, const Status endStatus)
	{
		Array<Endpoint*> result(endpoints.Size());
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				result[count++] = endpoints[i];
		}

		result.Resize(count);

		return result;
	}

	Array<Endpoint*> Comms::GetEndpoints(const EndDisp endDisp)
	{
		Array<Endpoint*> result(endpoints.Size());
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetDisposition() == endDisp)
				result[count++] = endpoints[i];

		result.Resize(count);

		return result;
	}

	UInt_64 Comms::GetEndpointsCount(const EndDisp endDisp, const Status endStatus)
	{
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetStatus() == endStatus)
				++count;
		}

		return count;
	}

	UInt_64 Comms::GetEndpointsCount(const EndDisp endDisp)
	{
		UInt_64 count = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
			if (endpoints[i]->GetDisposition() == endDisp)
				++count;

		return count;
	}

	UInt_64 Comms::GetMaxEndpoints() const
	{
		return maxEndpoints;
	}

	void Comms::SetBlocking(const bool blocking)
	{
		udp.SetBlocking(blocking);
	}

	bool Comms::IsBlocking() const
	{
		return udp.IsBlocking();
	}

	void Comms::SetMaxTimeout(const float seconds)
	{
		maxTimeout = seconds;
	}

	float Comms::GetMaxTimeout() const
	{
		return maxTimeout;
	}

    void Comms::SetResendRate(const float seconds)
    {
        resendRate = seconds;
    }

    float Comms::GetResendRate() const
    {
        return resendRate;
    }

	void Comms::SetConnectedCb(bool (*newCb)(Comms*, Endpoint*))
	{
		connectedCb = newCb;
	}

	void Comms::SetActiveCb(void (*newCb)(Comms*, Endpoint*))
	{
		activeCb = newCb;
	}

	void Comms::SetDisconnectedCb(void (*newCb)(Comms*, Endpoint*))
	{
		disconnectedCb = newCb;
	}

	void Comms::UpdateQueue(UInt_64 active)
	{
		UInt_64 slot = 0;

		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetStatus() == Status::IN_LOCAL_QUEUE)
			{
				if (active < maxEndpoints)
				{
					endpoints[i]->SetStatus(Status::ACTIVE);
					endpoints[i]->SetQueueSlot(0);

					Serializer payload(Endianness::LE);
					payload.Write(Status::ACTIVE);
					payload.Write(0);

					endpoints[i]->Send(false, true, false, internalSys, statusUpdateOp, payload);

					if (activeCb)
						activeCb(this, endpoints[i]);

					++active;
				}
				else
				{
					if (endpoints[i]->GetQueueSlot() != slot)
					{
						Serializer payload(Endianness::LE);
						payload.Write(Status::IN_REMOTE_QUEUE);
						payload.Write(slot);

						endpoints[i]->Send(false, true, false, internalSys, statusUpdateOp, payload);

						endpoints[i]->SetQueueSlot(slot++);
					}
					else
					{
						++slot;
					}
				}
			}
		}
	}

	void Comms::UpdateQueue()
	{
		UpdateQueue(GetEndpointsCount(EndDisp::ENDPOINT, Status::ACTIVE));
	}

	bool Comms::RemoveEndpoint(const EndDisp endDisp, const UInt_64 endHashId)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetDisposition() != endDisp)
				continue;

			if (endpoints[i]->GetHashId() == endHashId)
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	bool Comms::RemoveEndpoint(const Str_8& rAddress, const UInt_16 rPort)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i]->GetAddress() == rAddress && endpoints[i]->GetPort() == rPort)
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	bool Comms::RemoveEndpoint(const Endpoint* const end)
	{
		for (UInt_64 i = 0; i < endpoints.Size(); ++i)
		{
			if (endpoints[i] == end)
			{
				delete endpoints[i];

				if (i != endpoints.Size() - 1)
					endpoints.Swap(i, endpoints.End());

				endpoints.Pop();

				return true;
			}
		}

		return false;
	}

	void Comms::PollEndpoints()
	{
		UInt_64 i = 0;
		while (i < endpoints.Size())
		{
			endpoints[i]->Poll(delta);

			if (endpoints[i]->GetStatus() == Status::PENDING)
			{
				if (endpoints[i]->GetTimeout() >= maxTimeout)
				{
					EHS_LOG_INT("Info", 0, "Failed to connect to, \"" + endpoints[i]->GetAddress() + ":" + Str_8::FromNum(endpoints[i]->GetPort()) + "\".");

					delete endpoints[i];

					if (i != endpoints.Size() - 1)
						endpoints.Swap(i, endpoints.End());

					endpoints.Pop();

					continue;
				}
			}
			else
			{
				if (endpoints[i]->GetTimeout() >= maxTimeout)
				{
					EHS_LOG_INT("Info", 6, endpoints[i]->GetId() + " timed out.");

					if (disconnectedCb)
						disconnectedCb(this, endpoints[i]);

					delete endpoints[i];

					if (i != endpoints.Size() - 1)
						endpoints.Swap(i, endpoints.End());

					endpoints.Pop();

					UpdateQueue();

					continue;
				}

				Vector<Fragments>* frags = endpoints[i]->GetReceived();

                UInt_64 f = 0;
                while (f < frags->Size())
                {
                    if (!(*frags)[f].IsComplete())
                    {
                        ++f;
                        continue;
                    }

                    Packet packet = (*frags)[f].Combine();

                    CommsSystem* sys = GetSystem(packet.header.system);
                    if (!sys)
                    {
                        ++f;
                        continue;
                    }

                    sys->Execute(this, endpoints[i], packet.header.op, packet.payload);

                    frags->Swap(f, frags->End());
                    frags->Pop();
                }
			}

			++i;
		}
	}
}